import json
import typing
import warnings
from dataclasses import dataclass
from pathlib import Path

import yaml

from relari.eval.types import UUID, ToolCall

_SAFE_DICT = {k: v for k, v in typing.__dict__.items() if not k.startswith("__")}
_SAFE_DICT["UUID"] = UUID
_SAFE_DICT["ToolCall"] = ToolCall


@dataclass(frozen=True)
class DatasetField:
    name: str
    type: type
    description: str
    is_ground_truth: bool = False


@dataclass(frozen=True)
class DatasetManifest:
    name: str
    description: str
    format: str
    license: str
    fields: typing.Dict[str, DatasetField]


class Dataset:
    def __init__(
        self,
        dataset_path: typing.Union[str, Path],
        manifest_path: typing.Optional[typing.Union[str, Path]] = None,
    ) -> None:
        if isinstance(dataset_path, str):
            dataset_path = Path(dataset_path)
        assert dataset_path.exists(), f"Dataset {dataset_path} does not exist"
        if dataset_path.is_file() and manifest_path is None:
            manifest_path = dataset_path.parent / "manifest.yaml"
        elif dataset_path.is_dir():
            manifest_path = dataset_path / "manifest.yaml"
            dataset_path = dataset_path / "dataset.jsonl"
        else:
            manifest_path = None

        if dataset_path.suffix != ".jsonl":
            raise ValueError(f"Dataset file must be a .jsonl file, found {dataset_path.suffix}")
        # load jsonl dataset
        with open(dataset_path, "r") as json_file:
            self._data = [json.loads(x) for x in json_file.readlines()]
        self._manifest = self._load_or_infer_manifest(manifest_path)
        self._create_dynamic_properties()

    @classmethod
    def from_data(cls, data: typing.List[typing.Dict[str, typing.Any]]):
        dataset = cls.__new__(cls)
        dataset._data = data
        dataset._manifest = dataset._infer_manifest()
        dataset._create_dynamic_properties()
        return dataset

    def _load_or_infer_manifest(self, manifest_path: typing.Optional[Path]) -> DatasetManifest:
        if manifest_path is None or not manifest_path.exists():
            return self._infer_manifest()
        else:
            with open(manifest_path, "r") as manifest_file:
                manifest = yaml.safe_load(manifest_file)
                fields = {
                    field_name: DatasetField(
                        name=field_name,
                        type=eval(field_info["type"], _SAFE_DICT),
                        description=field_info.get("description", ""),
                        is_ground_truth=field_info.get("ground_truth", False),
                    )
                    for field_name, field_info in manifest["fields"].items()
                }
                return DatasetManifest(
                    name=manifest.get("name", ""),
                    description=manifest.get("description", ""),
                    format=manifest.get("format", ""),
                    license=manifest.get("license", ""),
                    fields=fields,
                )

    def _infer_manifest(self):
        fields = dict()
        for sample in self._data:
            for field_name, field_value in sample.items():
                dataset_field = DatasetField(
                    name=field_name,
                    type=type(field_value),
                    description="",
                    is_ground_truth=False,
                )
                if field_name not in fields:
                    fields[field_name] = dataset_field
                else:
                    if fields[field_name].type != dataset_field.type:
                        raise ValueError(
                            f"Field {field_name} has inconsistent types: {fields[field_name]['type']} and {type(field_value).__name__}"
                        )
        return DatasetManifest(
            name="",
            description="",
            format="",
            license="",
            fields=fields,
        )

    def _create_dynamic_properties(self):
        # Dynamically add a property for each field
        for field_name, field_info in self._manifest.fields.items():
            setattr(self, field_name, field_info)

    def filed_types(self, name: str) -> type:
        return getattr(self, name).type

    def save(self, file_path: typing.Union[str, Path], save_manifest: bool = False):
        if isinstance(file_path, str):
            file_path = Path(file_path)
        with open(file_path, "w") as json_file:
            for sample in self._data:
                json_file.write(json.dumps(sample) + "\n")

        if save_manifest:
            manifest_path = file_path.parent / "manifest.yaml"
            with open(manifest_path, "w") as manifest_file:
                manifest_file.write(yaml.dump(self._manifest.to_yaml()))

    @property
    def data(self):
        return self._data

    @property
    def name(self):
        return self._manifest.name

    @property
    def description(self):
        return self._manifest.description

    @property
    def format(self):
        return self._manifest.format

    @property
    def license(self):
        return self._manifest.license

    @property
    def fields(self) -> typing.List[DatasetField]:
        return list(self._manifest.fields.values())

    def __getitem__(self, key: str):
        return [x[key] for x in self._data]

    def __len__(self):
        return len(self._data)

    def filter(self, fcn: typing.Callable):
        self._data = [x for x in self._data if fcn(x)]
        return self



import json
import os
from typing import Optional, Dict
import urllib
import requests
from dotenv import load_dotenv

from relari.eval.manager import eval_manager

load_dotenv()


class RelariClient:
    def __init__(self, api_key: Optional[str] = None, url="https://api.relari.ai/v1/"):
        self._api_url = url
        if api_key is None:
            self.api_key = os.getenv("RELARI_API_KEY")
        else:
            self.api_key = api_key
        if self.api_key is None:
            raise ValueError(
                "Please set the environment variable RELARI_API_KEY or pass it as an argument."
            )

        self._headers = {"X-API-Key": self.api_key, "Content-Type": "application/json"}
        self.timeout = 10
        self.valid = self._validate()


    def status(self):
        try:
            response = requests.get(
                urllib.parse.urljoin(self._api_url, "status"),
                headers=self._headers,
                timeout=self.timeout,
            )
        except requests.exceptions.Timeout:
            exit("Request timed out while trying to validate API key")
        if response.status_code != 200:
            return False
        return True
        
    def _validate(self):
        try:
            response = requests.get(
                urllib.parse.urljoin(self._api_url, "auth/"),
                headers=self._headers,
                timeout=self.timeout,
            )
        except requests.exceptions.Timeout:
            exit("Request timed out while trying to validate API key")
        if response.status_code != 200:
            return False
        return True

    def start_remote_evaluation(self, pipeline_version: str):
        if eval_manager.is_running():
            raise ValueError("Cannot save while evaluation is running")
        payload = {
            "dataset": eval_manager.dataset.data,
            "version": pipeline_version,
            "pipeline": eval_manager.pipeline.asdict(),
            "results": eval_manager.evaluation.results,
            "metadata": eval_manager.metadata,
        }
        try:
            response = requests.post(
                urllib.parse.urljoin(self._api_url, "eval"),
                headers=self._headers,
                data=json.dumps(payload),
                timeout=self.timeout,
            )
        except requests.exceptions.Timeout:
            raise Exception("Request timed out while trying to start remote evaluation")
        if response.status_code != 200:
            raise Exception("Failed to start remote evaluation: " + response.text)
        print("Evaluation task submitted successfully")

    def run_metric(self, metric_name: str, args, **kwargs):
        # Version 1: metric_name, dataset, and sample_id
        if "dataset" in kwargs and "sample_id" in kwargs:
            endpoint = urllib.parse.urljoin(self._api_url, "metrics/dataset/")
            payload = {
                "metric": metric_name,
                "dataset": kwargs["dataset"],
                "id": kwargs["sample_id"],
                "kwargs": args,
            }
        # Version 2: metric_name and args
        elif isinstance(args, Dict):
            endpoint = urllib.parse.urljoin(self._api_url, "metrics/")
            payload = {
                "metric": metric_name,
                "kwargs": args,
            }
        else:
            raise ValueError("Invalid arguments provided to run_metric.")
        try:
            response = requests.post(
                endpoint,
                headers=self._headers,
                data=json.dumps(payload),
                timeout=self.timeout,
            )
        except requests.exceptions.Timeout:
            raise Exception("Request timed out while trying to run metric")
        if response.status_code != 200:
            raise Exception("Failed to run metric: " + response.text)
        return response.json()



from typing import Type, get_origin

def type_hint_to_str(type_hint: Type):
    if hasattr(type_hint, '__origin__'):  # Check if it's a generic type
        # Get the base type name (e.g., 'List' or 'Dict')
        base = type_hint.__origin__.__name__.title()
        # Recursively process the arguments (e.g., the contents of List, Dict, etc.)
        args = ", ".join(type_hint_to_str(arg) for arg in type_hint.__args__)
        return f"{base}[{args}]"
    elif hasattr(type_hint, '__name__'):
        return type_hint.__name__
    else:
        return repr(type_hint)


def instantiate_type(type_hint: Type):
    origin = get_origin(type_hint)
    # If the origin is None, it means type_hint is not a generic type
    # and we assume type_hint itself is directly instantiable
    if origin is None:
        origin = type_hint
    try:
        # This only works for types without required arguments in their __init__.
        instance = origin()
    except TypeError as e:
        # If instantiation fails, return an error message or raise a custom exception
        instance = None
    return instance



from dataclasses import dataclass, field, asdict
from typing import Dict, List, Optional, Type, Union, Iterable
from relari.eval.utils import type_hint_to_str
from relari.eval.dataset import DatasetField

InputType = Union[DatasetField, "Module"]

def _serialize_input_type(obj):
    if isinstance(obj, DatasetField):
        return {"__class__":obj.__class__.__name__, "name": obj.name}
    elif isinstance(obj, Module):
        return  {"__class__":obj.__class__.__name__, "name": obj.name}
    elif isinstance(obj, type):
        return type_hint_to_str(obj)
    elif isinstance(obj, (list, tuple)):
        return [serialize_input_type(x) for x in obj]
    else:
        raise TypeError(f"Object of type {type(obj).__name__} is not serializable")

@dataclass(frozen=True, eq=True)
class Module:
    name: str
    input: Union[Iterable[InputType], InputType, None]
    output: Type
    description: Optional[str] = field(default=None)

    def __post_init__(self):
        if self.name == "":
            raise ValueError(f"Module name cannot be empty")

    def asdict(self):
        return {
            "name": self.name,
            "input": _serialize_input_type(self.input),
            "output": type_hint_to_str(self.output),
            "description": self.description,
        }


@dataclass(frozen=True, eq=True)
class Tool:
    name: str
    args: Dict[str, Type]
    out_type: Type
    description: Optional[str] = field(default=None)

    def asdict(self):
        return asdict(self)


@dataclass(frozen=True, eq=True)
class AgentModule(Module):
    tools: Optional[List[Tool]] = field(default=None)

    def asdict(self):
        return asdict(self)



from dataclasses import dataclass
from typing import List, Set, Tuple

from relari.eval.dataset import Dataset, DatasetField
from relari.eval.modules import Module
from relari.eval.utils import type_hint_to_str


@dataclass(frozen=True)
class Graph:
    nodes: Set[str]
    edges: Set[Tuple[str, str]]
    dataset_edges: Set[Tuple[DatasetField, str]]


class Pipeline:
    def __init__(self, modules: List[Module], dataset: Dataset) -> None:
        self._modules = modules
        self._dataset = dataset
        self._graph = self._build_graph()

    @property
    def modules(self):
        return self._modules

    @property
    def dataset(self):
        return self._dataset

    def module_by_name(self, name: str) -> Module:
        for module in self._modules:
            if module.name == name:
                return module
        raise ValueError(f"Module {name} not found")

    def _validate_modules(self):
        names = set()
        for module in self._modules:
            if module.name in names:
                raise ValueError(f"Module {module.name} already exists")
            names.add(module.name)

    def _build_graph(self):
        nodes = {m.name for m in self._modules}
        edges = set()
        dataset_edges = set()
        for module in self._modules:
            if module.input is None:
                continue
            elif isinstance(module.input, Module):
                assert module in self._modules, f"Module {module.input.name} not found"
                edges.add((module.input.name, module.name))
            elif isinstance(module.input, DatasetField):
                assert (
                    module.input in self._dataset.fields
                ), f"Field {module.input.name} not found"
                dataset_edges.add((module.input, module.name))
            elif isinstance(module.input, (list, tuple)):
                for x in module.input:  # type: ignore
                    if isinstance(x, Module):
                        assert x in self._modules, f"Module {x.name} not found"
                        edges.add((x.name, module.name))
                    elif isinstance(x, DatasetField):
                        assert x in self._dataset.fields, f"Field {x.name} not found"
                        dataset_edges.add((x, module.name))
            else:
                raise ValueError(f"Invalid input type {module.input}")
        return Graph(nodes, edges, dataset_edges)

    def graph_repr(self, with_type_hints: bool = False):
        repr_str = "graph TD;\n"
        dataset_node_label = "Dataset"
        repr_str += f"    {dataset_node_label}(({dataset_node_label}));\n"
        for edge in self._graph.edges:
            start, end = edge
            if with_type_hints:
                type_hint = type_hint_to_str(self.module_by_name(start).output.type)
                repr_str += f'    {start}-->|"{type_hint}"|{end};\n'
            else:
                repr_str += f"    {start} --> {end};\n"
        for d_edge in self._graph.dataset_edges:
            dataset_field_name, end_node = d_edge[0].name, d_edge[1]
            # Adding the dataset edge with label
            repr_str += (
                f'    {dataset_node_label} -. "{dataset_field_name}" .-> {end_node};\n'
            )
        return repr_str

    def asdict(self):
        return {
            "modules": [m.asdict() for m in self._modules],
        }



import json
from collections import ChainMap
from functools import cached_property, lru_cache
from pathlib import Path
from typing import Dict, List, Optional

from relari.eval.modules import AgentModule
from relari.eval.pipeline import Pipeline
from relari.eval.utils import instantiate_type

TOOL_PREFIX = "_tool__"


class EvaluationResults:
    def __init__(self, pipeline: Optional[Pipeline] = None) -> None:
        if pipeline is None:
            self.results: List[Dict] = list()
        else:
            num_samples = len(pipeline.dataset.data)
            self.results: List[Dict] = [self._build_empty_samples(pipeline) for _ in range(num_samples)]

    def __len__(self):
        return len(self.results)

    def is_empty(self) -> bool:
        return not bool(self.results)

    def _build_empty_samples(self, pipeline: Pipeline):
        if pipeline is None:
            raise ValueError("Pipeline not set")
        empty_samples = dict()
        for module in pipeline.modules:
            empty_samples[module.name] = instantiate_type(module.output)
            if isinstance(module, AgentModule):
                empty_samples[f"{TOOL_PREFIX}{module.name}"] = list()
        return empty_samples

    def save(self, filepath: Path):
        assert filepath.suffix == ".jsonl", "File must be a JSONL file"
        assert self.results, "No samples to save"
        with open(filepath, "w") as f:
            for line in self.results:
                json_record = json.dumps(line, ensure_ascii=False)
                f.write(json_record + "\n")

    def load(self, filepath: Path):
        assert filepath.suffix == ".jsonl", "File must be a JSONL file"
        with open(filepath, "r") as f:
            self.results = [json.loads(line) for line in f]




from enum import Enum
from typing import Any, List, Optional


from relari.eval.dataset import Dataset, DatasetField
from relari.eval.pipeline import Pipeline
from relari.eval.result_types import TOOL_PREFIX, EvaluationResults


class LogMode(Enum):
    APPEND = 0
    REPLACE = 1


class EvaluationManager:
    def __init__(self):
        self._pipeline: Optional[Pipeline] = None
        self._eval_results: EvaluationResults = EvaluationResults()
        self._is_running: bool = False
        self._metadata = dict()

        self._idx = 0

    @property
    def samples(self) -> List[dict]:
        return self._eval_results.results

    @property
    def evaluation(self) -> EvaluationResults:
        return self._eval_results

    @property
    def metadata(self) -> dict:
        return self._metadata

    @property
    def pipeline(self) -> Pipeline:
        if self._pipeline is None:
            raise ValueError("Pipeline not set")
        return self._pipeline

    @property
    def dataset(self) -> Dataset:
        if self._pipeline is None:
            raise ValueError("Pipeline not set")
        if self._pipeline.dataset is None:
            raise ValueError("Dataset not set")
        return self._pipeline.dataset

    def set_pipeline(self, pipeline: Pipeline):
        self._pipeline = pipeline

    def set_metadata(self, metadata: dict):
        self._metadata = metadata

    def is_running(self) -> bool:
        return self._is_running

    def start_run(self, metadata: Optional[dict] = None):
        self._idx = 0
        self._is_running = True
        self._eval_results = EvaluationResults(self._pipeline)
        if metadata is not None:
            self.set_metadata(metadata)

    @property
    def curr_sample(self):
        if self._pipeline is None:
            raise ValueError("Pipeline not set")
        if self._idx >= len(self.dataset.data):
            self._is_running = False
            return None
        return self.dataset.data[self._idx]

    def next_sample(self):
        if self._pipeline is None:
            raise ValueError("Pipeline not set")
        if self._idx >= len(self.dataset.data):
            self._is_running = False
        else:
            self._idx += 1
        return self.curr_sample

    # Context manager
    def new_experiment(self, metadata: Optional[dict] = None):
        class ExperimentContext:
            def __init__(self, manager):
                self._manager = manager
                self._metadata = metadata

            def __enter__(self):
                # Initialize the session
                self._manager.start_run(self._metadata)
                return (
                    self  # Return the session object itself to be used as an iterator
                )

            def __exit__(self, exc_type, exc_val, exc_tb):
                # Clean up the session
                self._manager._is_running = False
                # Optional: Handle exceptions
                if exc_type:
                    print(f"An error occurred: {exc_val}")
                return False  # Propagate exceptions

            def __iter__(self):
                while (
                    self._manager.is_running() and self._manager.curr_sample is not None
                ):
                    yield self._manager.curr_sample
                    self._manager.next_sample()

        return ExperimentContext(self)

    # Logging results

    def log(
        self,
        module: str,
        value: Any,
        mode: LogMode = LogMode.REPLACE,
        **kwargs,
    ):
        # Make sure everything looks good
        if self._pipeline is None:
            raise ValueError("Pipeline not set")
        if not self._is_running:
            raise ValueError("Cannot log when not running")
        if module not in self._eval_results.results[self._idx]:
            raise ValueError(f"module {module} not found, review your pipeline")

        if kwargs and "tool_args" in kwargs:
            key = f"{TOOL_PREFIX}{module}"
            self._eval_results.results[self._idx][key].append(
                {"name": value, "kwargs": kwargs["tool_args"]}
            )
        else:
            if mode == LogMode.REPLACE:
                self._eval_results.results[self._idx][module] = value
            elif mode == LogMode.APPEND:
                if not isinstance(self._eval_results.results[self._idx][module], list):
                    if isinstance(value, list):
                        self._eval_results.results[self._idx][module].extend(value)
                    else:
                        self._eval_results.results[self._idx][module].append(value)
                else:
                    self._eval_results.results[self._idx][module].add(value)

eval_manager = EvaluationManager()
